

/*
   orxonox - the future of 3D-vertical-scrollers

   Copyright (C) 2004 orx

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   ### File Specific:
   main-programmer: Patrick Boenzli
   co-programmer: Christian Meyer
*/

#define DEBUG_SPECIAL_MODULE DEBUG_MODULE_WORLD

#include "world.h"

#include "shell_command.h"

#include "state.h"

#include "p_node.h"
#include "null_parent.h"
#include "pilot_node.h"
#include "track_node.h"
#include "world_entity.h"
#include "player.h"
#include "camera.h"
#include "environment.h"
#include "skysphere.h"
#include "skybox.h"
#include "satellite.h"
#include "test_entity.h"
#include "terrain.h"
#include "light.h"
#include "text_engine.h"
#include "load_param.h"
#include "shell.h"

#include "track_manager.h"
#include "garbage_collector.h"
#include "fast_factory.h"
#include "animation_player.h"
#include "particle_engine.h"
#include "graphics_engine.h"
#include "physics_engine.h"
#include "fields.h"

#include "md2Model.h"

#include "glmenu_imagescreen.h"
#include "list.h"
#include "game_loader.h"

#include "animation3d.h"

#include "substring.h"

#include "factory.h"

#include "projectile.h"
#include "event_handler.h"

#include "sound_engine.h"
#include "ogg_player.h"

#include "class_list.h"

#include "cd_engine.h"
#include "npc.h"

SHELL_COMMAND(speed, World, setSpeed);

using namespace std;

//! This creates a Factory to fabricate a World
CREATE_FACTORY(World);

World::World(const TiXmlElement* root)
{
  this->constuctorInit("", -1);
  this->path = NULL;

  this->loadParams(root);
}

/**
  *  create a new World

    This creates a new empty world!
*/
World::World (const char* name)
{
  this->path = NULL;
  this->constuctorInit(name, -1);
  //NullParent* np = NullParent::getInstance();
}

/**
 *  creates a new World...
 * @param worldID with this ID
*/
World::World (int worldID)
{
  this->path = NULL;
  this->constuctorInit(NULL, worldID);
}

/**
 *  remove the World from memory

    delete everything explicitly, that isn't contained in the parenting tree!
    things contained in the tree are deleted automaticaly
 */
World::~World ()
{
  delete this->shell;
  PRINTF(3)("World::~World() - deleting current world\n");

  // here everything that is alocated by the World is deleted
  delete this->entities;
  State::setWorldEntityList(NULL);


  delete LightManager::getInstance();
  delete TrackManager::getInstance();
  delete ParticleEngine::getInstance();
  delete AnimationPlayer::getInstance(); // this should be at the end of the unloading sequence.
  delete PhysicsEngine::getInstance();

  // external engines initialized by the orxonox-class get deleted
  SoundEngine::getInstance()->flushAllBuffers();
  SoundEngine::getInstance()->flushAllSources();
  FastFactory::flushAll(true);


  // erease everything that is left.
  delete NullParent::getInstance();


  // unload the resources !!
  ResourceManager::getInstance()->unload(this->music);
  ResourceManager::getInstance()->unloadAllByPriority(RP_LEVEL);
}

/**
 * initializes the world.
 * @param name the name of the world
 * @param worldID the ID of this world
 *
 * set all stuff here that is world generic and does not use to much memory
 * because the real init() function StoryEntity::init() will be called
 * shortly before start of the game.
 * since all worlds are initiated/referenced before they will be started.
 * NO LEVEL LOADING HERE - NEVER!
*/
void World::constuctorInit(const char* name, int worldID)
{
  this->setClassID(CL_WORLD, "World");

  this->setName(name);
  this->debugWorldNr = worldID;

  this->setSpeed(1.0);
  this->music = NULL;
}

/**
 * loads the parameters of a World from an XML-element
 * @param root the XML-element to load from
 */
void World::loadParams(const TiXmlElement* root)
{
  PRINTF(4)("Creating a World\n");

  LoadParam<World>(root, "identifier", this, &World::setStoryID)
    .describe("Sets the StoryID of this world");

  LoadParam<World>(root, "nextid", this, &World::setNextStoryID)
    .describe("Sets the ID of the next world");

  LoadParam<World>(root, "path", this, &World::setPath)
    .describe("The Filename of this World (relative from the data-dir)");
}

/**
 * this is executed just before load
 *
 * since the load function sometimes needs data, that has been initialized
 * before the load and after the proceeding storyentity has finished
*/
ErrorMessage World::preLoad()
{
  State::setWorldEntityList(this->entities = new tList<WorldEntity>());
  this->cycle = 0;

  /* init the world interface */
  this->shell = new Shell();

  LightManager::getInstance();
  NullParent::getInstance ();

  AnimationPlayer::getInstance(); // initializes the animationPlayer
  PhysicsEngine::getInstance();

  this->localCamera = new Camera();
  this->localCamera->setName ("World-Camera");

  State::setCamera(this->localCamera, this->localCamera->getTarget());

  GraphicsEngine::getInstance()->displayFPS(true);

  CDEngine::getInstance()->setEntityList( this->entities);
}


/**
 *  loads the World by initializing all resources, and set their default values.
*/
ErrorMessage World::load()
{
  PRINTF(3)("> Loading world: '%s'\n", getPath());
  TiXmlElement* element;
  GameLoader* loader = GameLoader::getInstance();

  if( getPath() == NULL)
    {
      PRINTF(1)("World has no path specified for loading");
      this->loadDebugWorld(this->getStoryID());
      return (ErrorMessage){213,"Path not specified","World::load()"};
    }

  TiXmlDocument* XMLDoc = new TiXmlDocument( path);
  // load the campaign document
  if( !XMLDoc->LoadFile())
  {
    // report an error
    PRINTF(1)("loading XML File: %s @ %d:%d\n", XMLDoc->ErrorDesc(), XMLDoc->ErrorRow(), XMLDoc->ErrorCol());
    delete XMLDoc;
    return (ErrorMessage){213,"XML File parsing error","World::load()"};
  }

  // check basic validity
  TiXmlElement* root = XMLDoc->RootElement();
  assert( root != NULL);

  if( root == NULL || root->Value() == NULL || strcmp( root->Value(), "WorldDataFile"))
    {
      // report an error
      PRINTF(1)("Specified XML File is not an orxonox world data file (WorldDataFile element missing)\n");
      delete XMLDoc;
      return (ErrorMessage){213,"Path not a WorldDataFile","World::load()"};
    }

  // load the parameters
  // name
  char* temp;
  const char* string = grabParameter( root, "name");
  if( string == NULL)
    {
      PRINTF(2)("World is missing a proper 'name'\n");
      string = "Unknown";
      temp = new char[strlen(string + 2)];
      strcpy( temp, string);
      this->setName(temp);
    }
  else
    {
      temp = new char[strlen(string + 2)];
      strcpy( temp, string);
      this->setName(temp);
    }

  ////////////////
  // LOADSCREEN //
  ////////////////
  element = root->FirstChildElement("LoadScreen");
  if (element == NULL)
    {
      PRINTF(2)("no LoadScreen specified, loading default\n");

      glmis->setBackgroundImage("pictures/load_screen.jpg");
      this->glmis->setMaximum(8);
      this->glmis->draw();
    }
  else
    {
      this->glmis->loadParams(element);
      this->glmis->draw();
    }
  this->glmis->draw();

  ////////////////////////
  // find WorldEntities //
  ////////////////////////

  element = root->FirstChildElement("WorldEntities");

  if( element == NULL)
    {
      PRINTF(1)("World is missing 'WorldEntities'\n");
    }
  else
    {
      element = element->FirstChildElement();
      // load Players/Objects/Whatever
      PRINTF(4)("Loading WorldEntities\n");
      while( element != NULL)
        {
          WorldEntity* created = dynamic_cast<WorldEntity*>( loader->fabricate( element));
          if( created != NULL) this->spawn( created);
          // if we load a 'Player' we use it as localPlayer
          //todo do this more elegant
          if( element->Value() != NULL && !strcmp( element->Value(), "Player"))
          {
            localPlayer = (Player*) created;
            CDEngine::getInstance()->setPlayer(localPlayer);
          }
          if( element->Value() != NULL && !strcmp( element->Value(), "SkyBox")) sky = (SkyBox*) created;
          if( element->Value() != NULL && !strcmp( element->Value(), "Terrain"))
          {
            terrain = (Terrain*) created;
            CDEngine::getInstance()->setTerrain(terrain);
          }
          element = element->NextSiblingElement();
          glmis->step(); //! @todo temporary
        }
      PRINTF(4)("Done loading WorldEntities\n");
    }

    //////////////////////////////
    // LOADING ADDITIONAL STUFF //
    //////////////////////////////

    LoadParam<LightManager>(root, "LightManager", LightManager::getInstance(), &LightManager::loadParams);

    LoadParam<ParticleEngine>(root, "ParticleEngine", ParticleEngine::getInstance(), &ParticleEngine::loadParams);
    LoadParam<PhysicsEngine>(root, "PhysicsEngine", PhysicsEngine::getInstance(), &PhysicsEngine::loadParams);

  // find Track
  element = root->FirstChildElement( "Track");
  if( element == NULL)
    {
      PRINTF(0)("World is missing a 'Track'\n");
    }
  else
    {
      //load track
      PRINTF(4)("Loading Track\n");

      TrackManager::getInstance()->loadParams( element);
      TrackManager::getInstance()->finalize();
    }

  // free the XML data

  delete XMLDoc;
  /* GENERIC LOADING PROCESS FINISHED */

  // bind input
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_UP);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_DOWN);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_LEFT);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_RIGHT);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_FIRE1);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_NEXT_WEAPON);
  EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_PREVIOUS_WEAPON);

  // bind camera
  //this->localCamera->bind (localPlayer);
 // this->localPlayer->addChild (this->localCamera);


  //        TrackManager::getInstance()->setBindSlave(env);
  PNode* tn = TrackManager::getInstance()->getTrackNode();
  tn->addChild(this->localPlayer);

  //localCamera->setParent(TrackNode::getInstance());
  tn->addChild(this->localCamera);
  localCamera->lookAt(tn);
  localCamera->setClipRegion(1, 10000.0);
  this->localPlayer->setParentMode(PNODE_ALL);
  Vector* cameraOffset = new Vector (0, 5, -10);
  TrackManager::getInstance()->condition(1, LEFTRIGHT, this->localPlayer);

  this->sky->setParent(this->localCamera);

  // initialize debug coord system
  objectList = glGenLists(1);
  glNewList (objectList, GL_COMPILE);

  //TrackManager::getInstance()->drawGraph(.01);
  //TrackManager::getInstance()->debug(2);
  glEndList();

  SoundEngine::getInstance()->setListener(this->localCamera);



  ////////////
  // STATIC //
  ////////////

  Gravity* test = new Gravity();

  // SYSTEM TRAILING THE PLAYER
  // Creating a Test Particle System

  //new PhysicsConnection(system, gravity);
  //    new PhysicsConnection(this->localPlayer, gravity);

//   TestEntity* testEntity = new TestEntity();
//   testEntity->setRelCoor(Vector(570, 10, -15));
//   testEntity->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
//   this->spawn(testEntity);

  TestEntity* testEntity2 = new TestEntity();
  testEntity2->setAnim(STAND);
  testEntity2->setRelCoor(Vector(2400.0, 10.0, -30.0));
  testEntity2->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  //testEntity2->setParent(this->localPlayer);
  this->spawn(testEntity2);

  TestEntity* testEntity3 = new TestEntity();
  testEntity3->setAnim(BOOM);
  testEntity3->setRelCoor(Vector(2450.0, 10.0, -40.0));
  testEntity3->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  this->spawn(testEntity3);

  TestEntity* testEntity4 = new TestEntity();
  testEntity4->setAnim(FLIP);
  testEntity4->setRelCoor(Vector(2500.0, 10.0, -22.0));
  testEntity4->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  this->spawn(testEntity4);

  TestEntity* testEntity5 = new TestEntity();
  testEntity5->setAnim(WAVE);
  testEntity5->setRelCoor(Vector(2420.0, 10.0, -50.0));
  testEntity5->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  this->spawn(testEntity5);

  TestEntity* testEntity6 = new TestEntity();
  testEntity6->setAnim(WAVE);
  testEntity6->setRelCoor(Vector(2420.0, 10.0, -20.0));
  testEntity6->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  this->spawn(testEntity6);

  TestEntity* testEntity7 = new TestEntity();
  testEntity7->setAnim(WAVE);
  testEntity7->setRelCoor(Vector(2500.0, 10.0, -50.0));
  testEntity7->setRelDir(Quaternion(M_PI, Vector(0, 1, 0)));
  this->spawn(testEntity7);



  PhysicsEngine::getInstance()->debug();



  for(int i = 0; i < 100; i++)
  {
    WorldEntity* tmp = new NPC();
    char npcChar[10];
    sprintf (npcChar, "NPC_%d", i);
    tmp->setName(npcChar);
    tmp->setAbsCoor(((float)rand()/RAND_MAX) * 5000, 50/*+ (float)rand()/RAND_MAX*20*/, ((float)rand()/RAND_MAX -.5) *100);
    this->spawn(tmp);


  }



  ClassList::debug();

  this->music = (OggPlayer*)ResourceManager::getInstance()->load("sound/00-luke_grey_-_hypermode.ogg", OGG, RP_LEVEL);
  music->playback();
}



/**
 * creates a debug world: only for experimental stuff
*/
void World::loadDebugWorld(int worldID)
{
  /*monitor progress*/
  this->glmis->step();
  // stuff beyond this point remains to be loaded properly

  // initializing the TrackManager
  TrackManager::getInstance()->addPointV(Vector(150, -35, 5));
  TrackManager::getInstance()->addPointV(Vector(200,-35, 5));
  TrackManager::getInstance()->addPointV(Vector(250, -35, 5));
  TrackManager::getInstance()->addPointV(Vector(320,-33,-.55));
  TrackManager::getInstance()->setDuration(1);
  TrackManager::getInstance()->setSavePoint();

  TrackManager::getInstance()->addPointV(Vector(410, 0, 0));
  TrackManager::getInstance()->addPointV(Vector(510, 20, -10));
  TrackManager::getInstance()->addPointV(Vector(550, 20, -10));
  TrackManager::getInstance()->addPointV(Vector(570, 20, -10));
  TrackManager::getInstance()->setDuration(2);

  TrackManager::getInstance()->forkS("testFork1,testFork2");
  TrackManager::getInstance()->workOnS("testFork1");
  TrackManager::getInstance()->addPointV(Vector(640, 25, -30));
  TrackManager::getInstance()->addPointV(Vector(700, 40, -120));
  TrackManager::getInstance()->addPointV(Vector(800, 50, -150));
  TrackManager::getInstance()->addPointV(Vector(900, 60, -100));
  TrackManager::getInstance()->addPointV(Vector(900, 60, -70));
  TrackManager::getInstance()->addPointV(Vector(990, 65, -15));
  TrackManager::getInstance()->addPointV(Vector(1050, 65, -10));
  TrackManager::getInstance()->addPointV(Vector(1100, 65, -20));
  TrackManager::getInstance()->setDuration(4);

  TrackManager::getInstance()->workOnS("testFork2");
  TrackManager::getInstance()->addPointV(Vector(640, 25, 20));
  TrackManager::getInstance()->addPointV(Vector(670, 50, 120));
  TrackManager::getInstance()->addPointV(Vector(700, 70, 80));
  TrackManager::getInstance()->addPointV(Vector(800, 70, 65));
  TrackManager::getInstance()->addPointV(Vector(850, 65, 65));
  TrackManager::getInstance()->addPointV(Vector(920, 35, 40));
  TrackManager::getInstance()->addPointV(Vector(945, 40, 40));
  TrackManager::getInstance()->addPointV(Vector(970, 24, 40));
  TrackManager::getInstance()->addPointV(Vector(1000, 40, -7));

  TrackManager::getInstance()->setDuration(4);


  TrackManager::getInstance()->joinS("testFork1,testFork2");

  TrackManager::getInstance()->addPointV(Vector(1200, 60, -50));
  TrackManager::getInstance()->addPointV(Vector(1300, 50, -50));
  TrackManager::getInstance()->addPointV(Vector(1400, 40, -50));
  TrackManager::getInstance()->addPointV(Vector(1500, 40, -60));
  TrackManager::getInstance()->addPointV(Vector(1600, 35, -55));
  TrackManager::getInstance()->addPointV(Vector(1700, 45, -40));
  TrackManager::getInstance()->addPointV(Vector(1750, 60, -40));
  TrackManager::getInstance()->addPointV(Vector(1770, 80, -40));
  TrackManager::getInstance()->addPointV(Vector(1800, 100, -40));
  TrackManager::getInstance()->setDuration(10);

  TrackManager::getInstance()->finalize();


  // LIGHT initialisation
  LightManager::getInstance()->setAmbientColor(.1,.1,.1);
//  LightManager::getInstance()->addLight();
  LightManager::getInstance()->debug();

  switch(this->debugWorldNr)
    {
      /*
        this loads the hard-coded debug world. this only for simplicity and will be
        removed by a reald world-loader, which interprets a world-file.
        if you want to add an own debug world, just add a case DEBUG_WORLD_[nr] and
        make whatever you want...
      */
    case DEBUG_WORLD_0:
      {
        LightManager::getInstance()->getLight()->setAbsCoor(-5.0, 10.0, -40.0);


        this->localPlayer = new Player ();
        this->localPlayer->setName ("player");
        this->spawn (this->localPlayer);
        this->localPlayer->setRelCoor(Vector(5,0,0));
        /*monitor progress*/
        this->glmis->step();


        EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_FIRE1);
        EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_NEXT_WEAPON);
        EventHandler::getInstance()->subscribe(this->localPlayer, ES_GAME, KeyMapper::PEV_PREVIOUS_WEAPON);

        /*
        Field* testField = new Gravity();
        testField->setMagnitude(10);
        new PhysicsConnection(this->localPlayer, testField);
        */

        // bind camera
        this->localCamera = new Camera();
        this->localCamera->setName ("camera");
        /*monitor progress*/
        this->glmis->step();


        // Create SkySphere
        this->sky = new Skysphere("pictures/sky-replace.jpg");
        this->sky->setName("SkySphere");
        this->spawn(this->sky);
        this->localCamera->addChild(this->sky);
        this->sky->setParentMode(PNODE_MOVEMENT);
        /*monitor progress*/
        this->glmis->step();


        terrain = new Terrain("worlds/newGround.obj");
        terrain->setRelCoor(Vector(0,-10,0));
        this->spawn(terrain);
        /*monitor progress*/
        this->glmis->step();

        this->pilotNode = new PilotNode();
        this->spawn(this->pilotNode);
        this->pilotNode->setAbsCoor(Vector(150, -35, 5));
        this->pilotNode->addChild(this->localPlayer);
        this->pilotNode->addChild(this->localCamera);
        this->localCamera->lookAt(this->localPlayer);

        EventHandler::getInstance()->subscribe(this->pilotNode, ES_GAME, KeyMapper::PEV_UP);
        EventHandler::getInstance()->subscribe(this->pilotNode, ES_GAME, KeyMapper::PEV_DOWN);
        EventHandler::getInstance()->subscribe(this->pilotNode, ES_GAME, KeyMapper::PEV_LEFT);
        EventHandler::getInstance()->subscribe(this->pilotNode, ES_GAME, KeyMapper::PEV_RIGHT);
        EventHandler::getInstance()->subscribe(this->pilotNode, ES_GAME, EV_MOUSE_MOTION);

        /*
        PNode* tn = TrackManager::getInstance()->getTrackNode();
        tn->addChild(this->localPlayer);
        this->localCamera->lookAt(tn);

        tn->addChild(this->localCamera);
        this->localPlayer->setParentMode(PNODE_ALL);
        TrackManager::getInstance()->condition(2, LEFTRIGHT, this->localPlayer);
        */
        this->glmis->step();
        break;
      }
    case DEBUG_WORLD_1:
      {

        break;
      }
    case DEBUG_WORLD_2:
      {

        break;
      }
    default:
      break;
    }
}

/**
 *  initializes a new World shortly before start
 *
 * this is the function, that will be loaded shortly before the world is
 * started
*/
ErrorMessage World::init()
{
  this->bPause = false;
  this->pilotNode = NULL;

  /* update the object position before game start - so there are no wrong coordinates used in the first processing */
  NullParent::getInstance()->update (0.001f);
  NullParent::getInstance()->update (0.001f);

}


/**
 *  starts the World
*/
ErrorMessage World::start()
{
  PRINTF(3)("World::start() - starting current World: nr %i\n", this->debugWorldNr);
  this->bQuitOrxonox = false;
  this->bQuitCurrentGame = false;
  this->mainLoop();
}

/**
 *  stops the world.

   This happens, when the player decides to end the Level.
*/
ErrorMessage World::stop()
{
  PRINTF(3)("World::stop() - got stop signal\n");
  this->bQuitCurrentGame = true;
}

/**
 *  pauses the Game
*/
ErrorMessage World::pause()
{
  this->isPaused = true;
}

/**
 *  ends the pause Phase
*/
ErrorMessage World::resume()
{
  this->isPaused = false;
}

/**
 *  destroys the World
*/
ErrorMessage World::destroy()
{

}

/**
 *  shows the loading screen
*/
void World::displayLoadScreen ()
{
  PRINTF(3)("World::displayLoadScreen - start\n");

  //GLMenuImageScreen*
  this->glmis = new GLMenuImageScreen();
  this->glmis->setMaximum(8);

  PRINTF(3)("World::displayLoadScreen - end\n");
}

/**
 *  removes the loadscreen, and changes over to the game

   @todo take out the delay
*/
void World::releaseLoadScreen ()
{
  PRINTF(3)("World::releaseLoadScreen - start\n");
  this->glmis->setValue(this->glmis->getMaximum());
  PRINTF(3)("World::releaseLoadScreen - end\n");
  delete this->glmis;
}


/**
 *  gets the list of entities from the world
 * @returns entity list
*/
tList<WorldEntity>* World::getEntities()
{
  return this->entities;
}


/**
 *  this returns the current game time
 * @returns elapsed game time
*/
double World::getGameTime()
{
  return this->gameTime;
}


/**
 *  function to put your own debug stuff into it. it can display informations about
   the current class/procedure
*/
void World::debug()
{
  PRINTF(0)("Printing out the List of alive WorldEntities:\n");
  tIterator<WorldEntity>* iterator = this->entities->getIterator();
  WorldEntity* entity = iterator->firstElement();
  while( entity != NULL)
  {
    PRINTF(0)("%s::%s\n", entity->getClassName(), entity->getName());
    entity = iterator->nextElement();
  }
  delete iterator;
}


/**
  \brief main loop of the world: executing all world relevant function

  in this loop we synchronize (if networked), handle input events, give the heart-beat to
  all other member-entities of the world (tick to player, enemies etc.), checking for
  collisions drawing everything to the screen.
*/
void World::mainLoop()
{
  this->lastFrame = SDL_GetTicks ();
  PRINTF(3)("World::mainLoop() - Entering main loop\n");

  while( !this->bQuitOrxonox && !this->bQuitCurrentGame) /* @todo implement pause */
    {
      ++this->cycle;
      PRINTF(4)("World::mainloop() - number of entities: %i\n", this->entities->getSize());
      // Network
      this->synchronize ();
      // Process input
      this->handleInput ();
      if( this->bQuitCurrentGame || this->bQuitOrxonox)
          break;
      // Process time
      this->tick ();
      // Process collision
      this->collide ();
      // Update the state
      this->update ();
      // Draw
      this->display ();
    }

  PRINTF(3)("World::mainLoop() - Exiting the main loop\n");
}


/**
 *  synchronize local data with remote data
*/
void World::synchronize ()
{
  // Get remote input
  // Update synchronizables
}


/**
 *  run all input processing

   the command node is the central input event dispatcher. the node uses the even-queue from
   sdl and has its own event-passing-queue.
*/
void World::handleInput ()
{
  // localinput
  //CommandNode* cn = Orxonox::getInstance()->getLocalInput();
  //cn->process();

  EventHandler::getInstance()->process();

  // remoteinput
}


/**
 *  advance the timeline

   this calculates the time used to process one frame (with all input handling, drawing, etc)
   the time is mesured in ms and passed to all world-entities and other classes that need
   a heart-beat.
*/
void World::tick ()
{
  Uint32 currentFrame = SDL_GetTicks();
  if(!this->bPause)
    {
      this->dt = currentFrame - this->lastFrame;

      if( this->dt > 10)
        {
          float fps = 1000/dt;

          // temporary, only for showing how fast the text-engine is
          char tmpChar[20];
          sprintf(tmpChar, "fps: %4.0f", fps);
        }
      else
        {
          /* the frame-rate is limited to 100 frames per second, all other things are for
             nothing.
          */
          PRINTF(3)("fps = 1000 - frame rate is adjusted\n");
          SDL_Delay(10-dt);
          this->dt = 10;
        }

      this->dtS = (float)this->dt / 1000.0 * this->speed;
      this->gameTime += this->dtS;

      tIterator<WorldEntity>* iterator = this->entities->getIterator();
      WorldEntity* entity = iterator->firstElement();
      while( entity != NULL)
        {
          entity->tick (this->dtS);
          entity = iterator->nextElement();
        }
      delete iterator;

      /* update tick the rest */
      TrackManager::getInstance()->tick(this->dtS);
      this->localCamera->tick(this->dtS);
      // tick the engines
      AnimationPlayer::getInstance()->tick(this->dtS);
//      if (this->cycle > 5)
        PhysicsEngine::getInstance()->tick(this->dtS);

      ParticleEngine::getInstance()->tick(this->dtS);
      GarbageCollector::getInstance()->tick(this->dtS);


      /** actualy the Graphics Engine should tick the world not the other way around...
         but since we like the things not too complicated we got it this way around
         until there is need or time to do it the other way around.
         @todo: GraphicsEngine ticks world: separation of processes and data...

        bensch: in my opinion the GraphicsEngine could draw the world, but not tick it,
         beceause graphics have nothing(or at least not much) to do with Motion.
      */
      GraphicsEngine::getInstance()->tick(this->dtS);
    }
  this->lastFrame = currentFrame;
}


/**
 *  this function gives the world a consistant state

   after ticking (updating the world state) this will give a constistant
   state to the whole system.
*/
void World::update()
{
  GarbageCollector::getInstance()->update();
  NullParent::getInstance()->update (this->dtS);
  GraphicsEngine::getInstance()->update(this->dtS);

  SoundEngine::getInstance()->update();
  //music->update();
}


void World::collide()
{
  CDEngine::getInstance()->checkCollisions();
}

/**
 *  render the current frame

   clear all buffers and draw the world
*/
void World::display ()
{
  // clear buffer
  glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  // set camera
  this->localCamera->apply ();
  // draw world
  this->draw();
  // draw HUD
  /** @todo draw HUD */
  // flip buffers
  GraphicsEngine::swapBuffers();
  //SDL_Surface* screen = Orxonox::getInstance()->getScreen ();
  //SDL_Flip (screen);
}


/**
 *  runs through all entities calling their draw() methods
 */
void World::draw ()
{
  /* draw entities */
  WorldEntity* entity;
  glLoadIdentity();
  tIterator<WorldEntity>* iterator = this->entities->getIterator();
  entity = iterator->firstElement();
  while( entity != NULL )
  {
    if( entity->isVisible() ) entity->draw();
    //entity->drawBVTree(2, 226);  // to draw the bounding boxes of the objects at level 2 for debug purp
    entity = iterator->nextElement();
  }
  delete iterator;

  glCallList (objectList);

  ParticleEngine::getInstance()->draw();

  GraphicsEngine::getInstance()->draw();
  //TextEngine::getInstance()->draw();
}

/**
 *  add and spawn a new entity to this world
 * @param entity to be added
*/
void World::spawn(WorldEntity* entity)
{
  this->entities->add (entity);
  entity->postSpawn ();
}


/**
 *  add and spawn a new entity to this world
 * @param entity to be added
 * @param absCoor At what coordinates to add this entity.
 * @param absDir In which direction should it look.
*/
void World::spawn(WorldEntity* entity, Vector* absCoor, Quaternion* absDir)
{
  this->entities->add (entity);

  entity->setAbsCoor (*absCoor);
  entity->setAbsDir (*absDir);

  entity->postSpawn ();
}


/**
 *  add and spawn a new entity to this world
 * @param entity to be added
 * @param entity to be added to (PNode)
 * @param At what relative  coordinates to add this entity.
 * @param In which relative direction should it look.
*/
void World::spawn(WorldEntity* entity, PNode* parentNode,
                  Vector* relCoor, Quaternion* relDir)
{
  NullParent::getInstance();
  if( parentNode != NULL)
    {
      parentNode->addChild (entity);

      entity->setRelCoor (*relCoor);
      entity->setRelDir (*relDir);

      this->entities->add (entity);

      entity->postSpawn ();
    }
}



/**
  \brief commands that the world must catch
  @returns false if not used by the world
*/
bool World::command(Command* cmd)
{
  if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW0)) this->localCamera->setViewMode(VIEW_NORMAL);
  else if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW1)) this->localCamera->setViewMode(VIEW_BEHIND);
  else if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW2)) this->localCamera->setViewMode(VIEW_FRONT);
  else if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW3)) this->localCamera->setViewMode(VIEW_LEFT);
  else if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW4)) this->localCamera->setViewMode(VIEW_RIGHT);
  else if( !strcmp( cmd->cmd, CONFIG_NAME_VIEW5)) this->localCamera->setViewMode(VIEW_TOP);
  return false;
}

void World::setPath( const char* name)
{
  if (this->path)
    delete this->path;
  if (ResourceManager::isFile(name))
  {
    this->path = new char[strlen(name)+1];
    strcpy(this->path, name);
  }
  else
    {
      this->path = new char[strlen(ResourceManager::getInstance()->getDataDir()) + strlen(name) +1];
      sprintf(this->path, "%s%s", ResourceManager::getInstance()->getDataDir(), name);
    }
}

const char* World::getPath( void)
{
  return path;
}
